@@ -2,17 +2,44 @@
 #include "fdiskP.h"
 #include "strutils.h"
 #include "carefulputc.h"
+#include "mangle.h"
+#include "jsonwrt.h"
+#include "fileutils.h"
+
+#ifdef FUZZ_TARGET
+#include "fuzz.h"
+#endif
 
 /**
  * SECTION: script
  * @title: Script
- * @short_description: text based sfdisk compatible description of partition table
+ * @short_description: complex way to create and dump partition table
  *
- * The libfdisk scripts are based on original sfdisk script (dumps).  Each
+ * This interface can be used to compose in-memory partition table with all details,
+ * write all partition table description to human readable text file, read it
+ * from the file, and apply the script to on-disk label.
+ *
+ * The libfdisk scripts are based on original sfdisk script (dumps). Each
  * script has two parts: script headers and partition table entries
- * (partitions).
+ * (partitions). The script is possible to dump in JSON too (read JSON is not
+ * implemented yet).
  *
  * For more details about script format see sfdisk man page.
+ *
+ * There are four ways how to build the script:
+ *
+ * - read the current on-disk partition table by fdisk_script_read_context())
+ * - read it from text file by fdisk_script_read_file()
+ * - read it interactively from user by fdisk_script_read_line() and fdisk_script_set_fgets()
+ * - manually in code by fdisk_script_set_header() and fdisk_script_set_table()
+ *
+ * The read functions fdisk_script_read_context() and fdisk_script_read_file()
+ * creates always a new script partition table. The table (see
+ * fdisk_script_get_table()) is possible to modify by standard
+ * fdisk_table_...() functions and then apply by fdisk_apply_script().
+ *
+ * Note that script API is fully non-interactive and forces libfdisk to not use
+ * standard dialog driven partitioning as we have in fdisk(8).
  */
 
 /* script header (e.g. unit: sectors) */
@@ -36,11 +63,12 @@ struct fdisk_script {
 	size_t			nlines;
 	struct fdisk_label	*label;
 
+	unsigned long		sector_size;		/* as defined by script */
+
 	unsigned int		json : 1,		/* JSON output */
 				force_label : 1;	/* label: <name> specified */
 };
 
-
 static void fdisk_script_free_header(struct fdisk_scriptheader *fi)
 {
 	if (!fi)
@@ -75,12 +103,6 @@ struct fdisk_script *fdisk_new_script(struct fdisk_context *cxt)
 	dp->cxt = cxt;
 	fdisk_ref_context(cxt);
 
-	dp->table = fdisk_new_table();
-	if (!dp->table) {
-		fdisk_unref_script(dp);
-		return NULL;
-	}
-
 	INIT_LIST_HEAD(&dp->headers);
 	return dp;
 }
@@ -147,8 +169,9 @@ static void fdisk_reset_script(struct fdisk_script *dp)
 	assert(dp);
 
 	DBG(SCRIPT, ul_debugobj(dp, "reset"));
-	fdisk_unref_table(dp->table);
-	dp->table = NULL;
+
+	if (dp->table)
+		fdisk_reset_table(dp->table);
 
 	while (!list_empty(&dp->headers)) {
 		struct fdisk_scriptheader *fi = list_entry(dp->headers.next,
@@ -174,6 +197,7 @@ void fdisk_unref_script(struct fdisk_script *dp)
 	if (dp->refcount <= 0) {
 		fdisk_reset_script(dp);
 		fdisk_unref_context(dp->cxt);
+		fdisk_unref_table(dp->table);
 		DBG(SCRIPT, ul_debugobj(dp, "free script"));
 		free(dp);
 	}
@@ -240,7 +264,6 @@ const char *fdisk_script_get_header(struct fdisk_script *dp, const char *name)
 	return fi ? fi->data : NULL;
 }
 
-
 /**
  * fdisk_script_set_header:
  * @dp: script instance
@@ -253,8 +276,8 @@ const char *fdisk_script_get_header(struct fdisk_script *dp, const char *name)
  * If no @data is specified then the header is removed. If header does not exist
  * and @data is specified then a new header is added.
  *
- * Note that libfdisk allows to specify arbitrary custom header, the default
- * build-in headers are "unit" and "label", and some label specific headers
+ * Note that libfdisk can be used to specify arbitrary custom header, the default
+ * built-in headers are "unit" and "label", and some label specific headers
  * (for example "uuid" and "name" for GPT).
  *
  * Returns: 0 on success, <0 on error
@@ -281,6 +304,8 @@ int fdisk_script_set_header(struct fdisk_script *dp,
 	}
 
 	if (!fi) {
+		int rc;
+
 		DBG(SCRIPT, ul_debugobj(dp, "setting new header %s='%s'", name, data));
 
 		/* new header */
@@ -288,11 +313,13 @@ int fdisk_script_set_header(struct fdisk_script *dp,
 		if (!fi)
 			return -ENOMEM;
 		INIT_LIST_HEAD(&fi->headers);
-		fi->name = strdup(name);
-		fi->data = strdup(data);
-		if (!fi->data || !fi->name) {
+
+		rc = strdup_to_struct_member(fi, name, name);
+		if (!rc)
+			rc = strdup_to_struct_member(fi, data, data);
+		if (rc) {
 			fdisk_script_free_header(fi);
-			return -ENOMEM;
+			return rc;
 		}
 		list_add_tail(&fi->headers, &dp->headers);
 	} else {
@@ -317,16 +344,60 @@ int fdisk_script_set_header(struct fdisk_script *dp,
  * fdisk_script_get_table:
  * @dp: script
  *
- * The table (container with partitions) is possible to create by
- * fdisk_script_read_context() or fdisk_script_read_file(), otherwise
- * this function returns NULL.
+ * The table represents partitions holded by the script. The table is possible to
+ * fill by fdisk_script_read_context() or fdisk_script_read_file(). All the "read"
+ * functions remove old partitions from the table. See also fdisk_script_set_table().
  *
- * Returns: NULL or script.
+ * Returns: NULL or script table.
  */
 struct fdisk_table *fdisk_script_get_table(struct fdisk_script *dp)
 {
 	assert(dp);
-	return dp ? dp->table : NULL;
+
+	if (!dp->table)
+		/*
+		 * Make sure user has access to the same table as script. If
+		 * there is no table then create a new one and reuse it later.
+		 */
+		dp->table = fdisk_new_table();
+
+	return dp->table;
+}
+
+/**
+ * fdisk_script_set_table:
+ * @dp: script
+ * @tb: table
+ *
+ * Replaces table used by script and creates a new reference to @tb. This
+ * function can be used to generate a new script table independently on the current
+ * context and without any file reading.
+ *
+ * This is useful for example to create partition table with the same basic
+ * settings (e.g. label-id, ...) but with different partitions -- just call
+ * fdisk_script_read_context() to get current settings and then
+ * fdisk_script_set_table() to set a different layout.
+ *
+ * If @tb is NULL then the current script table is unreferenced.
+ *
+ * Note that script read_ functions (e.g. fdisk_script_read_context()) create
+ * always a new script table.
+ *
+ * Returns: 0 on success, <0 on error
+ *
+ * Since: 2.35
+ */
+int fdisk_script_set_table(struct fdisk_script *dp, struct fdisk_table *tb)
+{
+	if (!dp)
+		return -EINVAL;
+
+	fdisk_ref_table(tb);
+	fdisk_unref_table(dp->table);
+	dp->table = tb;
+
+	DBG(SCRIPT, ul_debugobj(dp, "table replaced"));
+	return 0;
 }
 
 static struct fdisk_label *script_get_label(struct fdisk_script *dp)
@@ -377,7 +448,7 @@ int fdisk_script_has_force_label(struct fdisk_script *dp)
  * @cxt: context
  *
  * Reads data from the @cxt context (on disk partition table) into the script.
- * If the context is no specified than defaults to context used for fdisk_new_script().
+ * If the context is not specified then defaults to context used for fdisk_new_script().
  *
  * Return: 0 on success, <0 on error.
  */
@@ -386,6 +457,7 @@ int fdisk_script_read_context(struct fdisk_script *dp, struct fdisk_context *cxt
 	struct fdisk_label *lb;
 	int rc;
 	char *p = NULL;
+	char buf[64];
 
 	if (!dp || (!cxt && !dp->cxt))
 		return -EINVAL;
@@ -400,7 +472,7 @@ int fdisk_script_read_context(struct fdisk_script *dp, struct fdisk_context *cxt
 	if (!lb)
 		return -EINVAL;
 
-	/* allocate and fill new table */
+	/* allocate (if not yet) and fill table */
 	rc = fdisk_get_partitions(cxt, &dp->table);
 	if (rc)
 		return rc;
@@ -418,8 +490,7 @@ int fdisk_script_read_context(struct fdisk_script *dp, struct fdisk_context *cxt
 		rc = fdisk_script_set_header(dp, "unit", "sectors");
 
 	if (!rc && fdisk_is_label(cxt, GPT)) {
-		struct fdisk_labelitem item;
-		char buf[64];
+		struct fdisk_labelitem item = FDISK_LABELITEM_INIT;
 
 		/* first-lba */
 		rc = fdisk_get_disklabel_item(cxt, GPT_LABELITEM_FIRSTLBA, &item);
@@ -446,6 +517,16 @@ int fdisk_script_read_context(struct fdisk_script *dp, struct fdisk_context *cxt
 		}
 	}
 
+	if (!rc && fdisk_get_grain_size(cxt) != 2048 * 512) {
+		snprintf(buf, sizeof(buf), "%lu", fdisk_get_grain_size(cxt));
+		rc = fdisk_script_set_header(dp, "grain", buf);
+	}
+
+	if (!rc) {
+		snprintf(buf, sizeof(buf), "%lu", fdisk_get_sector_size(cxt));
+		rc = fdisk_script_set_header(dp, "sector-size", buf);
+	}
+
 	DBG(SCRIPT, ul_debugobj(dp, "read context done [rc=%d]", rc));
 	return rc;
 }
@@ -467,32 +548,23 @@ int fdisk_script_enable_json(struct fdisk_script *dp, int json)
 	return 0;
 }
 
-static void fput_indent(int indent, FILE *f)
-{
-	int i;
-
-	for (i = 0; i <= indent; i++)
-		fputs("   ", f);
-}
-
 static int write_file_json(struct fdisk_script *dp, FILE *f)
 {
 	struct list_head *h;
 	struct fdisk_partition *pa;
 	struct fdisk_iter itr;
 	const char *devname = NULL;
-	int ct = 0, indent = 0;
+	struct ul_jsonwrt json;
 
 	assert(dp);
 	assert(f);
 
 	DBG(SCRIPT, ul_debugobj(dp, "writing json dump to file"));
 
-	fputs("{\n", f);
+	ul_jsonwrt_init(&json, f, 0);
+	ul_jsonwrt_root_open(&json);
 
-	fput_indent(indent, f);
-	fputs("\"partitiontable\": {\n", f);
-	indent++;
+	ul_jsonwrt_object_open(&json, "partitiontable");
 
 	/* script headers */
 	list_for_each(h, &dp->headers) {
@@ -506,92 +578,81 @@ static int write_file_json(struct fdisk_script *dp, FILE *f)
 		} else if (strcmp(name, "last-lba") == 0) {
 			name = "lastlba";
 			num = 1;
+		} else if (strcmp(name, "sector-size") == 0) {
+			name = "sectorsize";
+			num = 1;
 		} else if (strcmp(name, "label-id") == 0)
 			name = "id";
 
-		fput_indent(indent, f);
-		fputs_quoted_lower(name, f);
-		fputs(": ", f);
-		if (!num)
-			fputs_quoted(fi->data, f);
+		if (num)
+			ul_jsonwrt_value_raw(&json, name, fi->data);
 		else
-			fputs(fi->data, f);
-		if (!dp->table && fi == list_last_entry(&dp->headers, struct fdisk_scriptheader, headers))
-			fputc('\n', f);
-		else
-			fputs(",\n", f);
+			ul_jsonwrt_value_s(&json, name, fi->data);
 
 		if (strcmp(name, "device") == 0)
 			devname = fi->data;
 	}
 
 
-	if (!dp->table) {
+	if (!dp->table || fdisk_table_is_empty(dp->table)) {
 		DBG(SCRIPT, ul_debugobj(dp, "script table empty"));
 		goto done;
 	}
 
 	DBG(SCRIPT, ul_debugobj(dp, "%zu entries", fdisk_table_get_nents(dp->table)));
 
-	fput_indent(indent, f);
-	fputs("\"partitions\": [\n", f);
-	indent++;
+	ul_jsonwrt_array_open(&json, "partitions");
 
 	fdisk_reset_iter(&itr, FDISK_ITER_FORWARD);
 	while (fdisk_table_next_partition(dp->table, &itr, &pa) == 0) {
 		char *p = NULL;
 
-		ct++;
-		fput_indent(indent, f);
-		fputc('{', f);
+		ul_jsonwrt_object_open(&json, NULL);
 		if (devname)
 			p = fdisk_partname(devname, pa->partno + 1);
 		if (p) {
 			DBG(SCRIPT, ul_debugobj(dp, "write %s entry", p));
-			fputs("\"node\": ", f);
-			fputs_quoted(p, f);
+			ul_jsonwrt_value_s(&json, "node", p);
+			free(p);
 		}
 
 		if (fdisk_partition_has_start(pa))
-			fprintf(f, ", \"start\": %ju", (uintmax_t)pa->start);
+			ul_jsonwrt_value_u64(&json, "start", (uintmax_t)pa->start);
+
 		if (fdisk_partition_has_size(pa))
-			fprintf(f, ", \"size\": %ju", (uintmax_t)pa->size);
+			ul_jsonwrt_value_u64(&json, "size", (uintmax_t)pa->size);
 
 		if (pa->type && fdisk_parttype_get_string(pa->type))
-			fprintf(f, ", \"type\": \"%s\"", fdisk_parttype_get_string(pa->type));
-		else if (pa->type)
-			fprintf(f, ", \"type\": \"%x\"", fdisk_parttype_get_code(pa->type));
+			ul_jsonwrt_value_s(&json, "type", fdisk_parttype_get_string(pa->type));
+
+		else if (pa->type) {
+			ul_jsonwrt_value_open(&json, "type");
+			fprintf(f, "\"%x\"", fdisk_parttype_get_code(pa->type));
+			ul_jsonwrt_value_close(&json);
+		}
 
 		if (pa->uuid)
-			fprintf(f, ", \"uuid\": \"%s\"", pa->uuid);
-		if (pa->name && *pa->name) {
-			fputs(", \"name\": ", f),
-			fputs_quoted(pa->name, f);
-		}
+			ul_jsonwrt_value_s(&json, "uuid", pa->uuid);
+		if (pa->name && *pa->name)
+			ul_jsonwrt_value_s(&json, "name", pa->name);
 
 		/* for MBR attr=80 means bootable */
 		if (pa->attrs) {
 			struct fdisk_label *lb = script_get_label(dp);
 
 			if (!lb || fdisk_label_get_type(lb) != FDISK_DISKLABEL_DOS)
-				fprintf(f, ", \"attrs\": \"%s\"", pa->attrs);
+				ul_jsonwrt_value_s(&json, "attrs", pa->attrs);
 		}
-		if (fdisk_partition_is_bootable(pa))
-			fprintf(f, ", \"bootable\": true");
 
-		if ((size_t)ct < fdisk_table_get_nents(dp->table))
-			fputs("},\n", f);
-		else
-			fputs("}\n", f);
+		if (fdisk_partition_is_bootable(pa))
+			ul_jsonwrt_value_boolean(&json, "bootable", 1);
+		ul_jsonwrt_object_close(&json);
 	}
 
-	indent--;
-	fput_indent(indent, f);
-	fputs("]\n", f);
+	ul_jsonwrt_array_close(&json);
 done:
-	indent--;
-	fput_indent(indent, f);
-	fputs("}\n}\n", f);
+	ul_jsonwrt_object_close(&json);
+	ul_jsonwrt_root_close(&json);
 
 	DBG(SCRIPT, ul_debugobj(dp, "write script done"));
 	return 0;
@@ -617,7 +678,7 @@ static int write_file_sfdisk(struct fdisk_script *dp, FILE *f)
 			devname = fi->data;
 	}
 
-	if (!dp->table) {
+	if (!dp->table || fdisk_table_is_empty(dp->table)) {
 		DBG(SCRIPT, ul_debugobj(dp, "script table empty"));
 		return 0;
 	}
@@ -635,6 +696,7 @@ static int write_file_sfdisk(struct fdisk_script *dp, FILE *f)
 		if (p) {
 			DBG(SCRIPT, ul_debugobj(dp, "write %s entry", p));
 			fprintf(f, "%s :", p);
+			free(p);
 		} else
 			fprintf(f, "%zu :", pa->partno + 1);
 
@@ -650,8 +712,10 @@ static int write_file_sfdisk(struct fdisk_script *dp, FILE *f)
 
 		if (pa->uuid)
 			fprintf(f, ", uuid=%s", pa->uuid);
-		if (pa->name && *pa->name)
-			fprintf(f, ", name=\"%s\"", pa->name);
+		if (pa->name && *pa->name) {
+			fputs(", name=", f);
+			fputs_quoted(pa->name, f);
+		}
 
 		/* for MBR attr=80 means bootable */
 		if (pa->attrs) {
@@ -701,8 +765,12 @@ static inline int is_header_line(const char *s)
 /* parses "<name>: value", note modifies @s*/
 static int parse_line_header(struct fdisk_script *dp, char *s)
 {
-	int rc = -EINVAL;
+	size_t i;
 	char *name, *value;
+	static const char *supported[] = {
+		"label", "unit", "label-id", "device", "grain",
+		"first-lba", "last-lba", "table-length", "sector-size"
+	};
 
 	DBG(SCRIPT, ul_debugobj(dp, "   parse header '%s'", s));
 
@@ -712,7 +780,7 @@ static int parse_line_header(struct fdisk_script *dp, char *s)
 	name = s;
 	value = strchr(s, ':');
 	if (!value)
-		goto done;
+		return -EINVAL;
 	*value = '\0';
 	value++;
 
@@ -721,33 +789,75 @@ static int parse_line_header(struct fdisk_script *dp, char *s)
 	ltrim_whitespace((unsigned char *) value);
 	rtrim_whitespace((unsigned char *) value);
 
+	if (!*name || !*value)
+		return -EINVAL;
+
+	/* check header name */
+	for (i = 0; i < ARRAY_SIZE(supported); i++) {
+		if (strcmp(name, supported[i]) == 0)
+			break;
+	}
+	if (i == ARRAY_SIZE(supported))
+		return -ENOTSUP;
+
+	/* header specific actions */
 	if (strcmp(name, "label") == 0) {
 		if (dp->cxt && !fdisk_get_label(dp->cxt, value))
-			goto done;			/* unknown label name */
+			return -EINVAL;			/* unknown label name */
 		dp->force_label = 1;
+
+	} else if (strcmp(name, "sector-size") == 0) {
+		uint64_t x = 0;
+
+		if (ul_strtou64(value, &x, 10) != 0)
+			return -EINVAL;
+		if (x > ULONG_MAX || x % 512)
+			return -ERANGE;
+		dp->sector_size = (unsigned long) x;
+
+		if (dp->cxt && dp->sector_size && dp->cxt->sector_size
+		    && dp->sector_size != dp->cxt->sector_size)
+			fdisk_warnx(dp->cxt, _("The script and device sector size differ; the sizes will be recalculated to match the device."));
+
 	} else if (strcmp(name, "unit") == 0) {
 		if (strcmp(value, "sectors") != 0)
-			goto done;			/* only "sectors" supported */
-	} else if (strcmp(name, "label-id") == 0
-		   || strcmp(name, "device") == 0
-		   || strcmp(name, "first-lba") == 0
-		   || strcmp(name, "last-lba") == 0
-		   || strcmp(name, "table-length") == 0) {
-		;					/* whatever is possible */
-	} else
-		goto done;				/* unknown header */
+			return -EINVAL;			/* only "sectors" supported */
 
-	if (*name && *value)
-		rc = fdisk_script_set_header(dp, name, value);
-done:
-	if (rc)
-		DBG(SCRIPT, ul_debugobj(dp, "header parse error: "
-				"[rc=%d, name='%s', value='%s']",
-				rc, name, value));
-	return rc;
+	}
 
+	return fdisk_script_set_header(dp, name, value);
 }
 
+static int partno_from_devname(char *s)
+{
+	intmax_t num;
+	size_t sz;
+	char *end, *p;
+
+	if (!s || !*s)
+		return -1;
+
+	sz = rtrim_whitespace((unsigned char *)s);
+	end = p = s + sz;
+
+	while (p > s && isdigit(*(p - 1)))
+		p--;
+	if (p == end)
+		return -1;
+	end = NULL;
+	errno = 0;
+	num = strtol(p, &end, 10);
+	if (errno || !end || p == end)
+		return -1;
+
+	if (num < INT32_MIN || num > INT32_MAX) {
+		errno = ERANGE;
+		return -1;
+	}
+	return num - 1;
+}
+
+
 /* returns zero terminated string with next token and @str is updated */
 static char *next_token(char **str)
 {
@@ -812,23 +922,37 @@ static char *next_token(char **str)
 	return tk_begin;
 }
 
-static int next_number(char **s, uint64_t *num, int *power)
+/*
+ * "[-]<,;>"
+ * "[ ]<,;>"
+ * "- <value>"
+ */
+static int is_default_value(char **str)
 {
-	char *tk;
-	int rc = -EINVAL;
+	char *p = (char *) skip_blank(*str);
+	int blank = 0;
 
-	assert(num);
-	assert(s);
+	if (*p == '-') {
+		char *x = ++p;
+		p = (char *) skip_blank(x);
+		blank = x < p;			/* "- " */
+	}
 
-	tk = next_token(s);
-	if (tk)
-		rc = parse_size(tk, (uintmax_t *) num, power);
-	return rc;
+	if (*p == ';' || *p == ',') {
+		*str = ++p;
+		return 1;
+	}
+	if (*p == '\0' || blank) {
+		*str = p;
+		return 1;
+	}
+
+	return 0;
 }
 
 static int next_string(char **s, char **str)
 {
-	char *tk;
+	char *tk, *p = NULL;
 	int rc = -EINVAL;
 
 	assert(s);
@@ -836,31 +960,167 @@ static int next_string(char **s, char **str)
 
 	tk = next_token(s);
 	if (tk) {
-		*str = strdup(tk);
-		rc = !*str ? -ENOMEM : 0;
+		p = strdup(tk);
+		rc = p ? 0 : -ENOMEM;
 	}
+
+	*str = p;
 	return rc;
 }
 
-static int partno_from_devname(char *s)
+static int skip_optional_sign(char **str)
 {
-	int pno;
-	size_t sz;
-	char *end, *p;
+	char *p = (char *) skip_blank(*str);
 
-	sz = rtrim_whitespace((unsigned char *)s);
-	p = s + sz - 1;
-
-	while (p > s && isdigit(*(p - 1)))
-		p--;
-
-	errno = 0;
-	pno = strtol(p, &end, 10);
-	if (errno || !end || p == end)
-		return -1;
-	return pno - 1;
+	if (*p == '-' || *p == '+') {
+		*str = p+1;
+		return *p;
+	}
+	return 0;
 }
 
+static int recount_script2device_sectors(struct fdisk_script *dp, uint64_t *num)
+{
+	if (!dp->cxt ||
+	    !dp->sector_size ||
+	    !dp->cxt->sector_size)
+		return 0;
+
+	if (dp->sector_size > dp->cxt->sector_size)
+		 *num *= (dp->sector_size / dp->cxt->sector_size);
+
+	else if (dp->sector_size < dp->cxt->sector_size) {
+		uint64_t x = dp->cxt->sector_size / dp->sector_size;
+
+		if (*num % x)
+			return -EINVAL;
+		*num /= x;
+	}
+
+	return 0;
+}
+
+static int parse_start_value(struct fdisk_script *dp, struct fdisk_partition *pa, char **str)
+{
+	char *tk;
+	int rc = 0;
+
+	assert(str);
+
+	if (is_default_value(str)) {
+		fdisk_partition_start_follow_default(pa, 1);
+		return 0;
+	}
+
+	tk = next_token(str);
+	if (!tk)
+		return -EINVAL;
+
+	if (strcmp(tk, "+") == 0) {
+		fdisk_partition_start_follow_default(pa, 1);
+		pa->movestart = FDISK_MOVE_DOWN;
+	} else {
+		int pow = 0, sign = skip_optional_sign(&tk);
+		uint64_t num;
+
+		rc = parse_size(tk, (uintmax_t *) &num, &pow);
+		if (!rc) {
+			if (pow) {	/* specified as <num><suffix> */
+				if (!dp->cxt->sector_size) {
+					rc = -EINVAL;
+					goto done;
+				}
+				num /= dp->cxt->sector_size;
+			} else {
+				rc = recount_script2device_sectors(dp, &num);
+				if (rc) {
+					fdisk_warnx(dp->cxt, _("Can't recalculate partition start to the device sectors"));
+					goto done;
+				}
+			}
+
+			fdisk_partition_set_start(pa, num);
+
+			pa->movestart = sign == '-' ? FDISK_MOVE_DOWN :
+					sign == '+' ? FDISK_MOVE_UP :
+						      FDISK_MOVE_NONE;
+		}
+		fdisk_partition_start_follow_default(pa, 0);
+	}
+
+done:
+	DBG(SCRIPT, ul_debugobj(dp, "  start parse result: rc=%d, move=%s, start=%ju, default=%s",
+				rc, pa->movestart == FDISK_MOVE_DOWN ? "down" :
+				    pa->movestart == FDISK_MOVE_UP ? "up" : "none",
+				    pa->start,
+				    pa->start_follow_default ? "on" : "off"));
+	return rc;
+}
+
+static int parse_size_value(struct fdisk_script *dp, struct fdisk_partition *pa, char **str)
+{
+	char *tk;
+	int rc = 0;
+
+	if (is_default_value(str)) {
+		fdisk_partition_end_follow_default(pa, 1);
+		return 0;
+	}
+
+	tk = next_token(str);
+	if (!tk)
+		return -EINVAL;
+
+	if (strcmp(tk, "+") == 0) {
+		fdisk_partition_end_follow_default(pa, 1);
+		pa->resize = FDISK_RESIZE_ENLARGE;
+	} else {
+		/* '[+-]<number>[<suffix] */
+		int pow = 0, sign = skip_optional_sign(&tk);
+		uint64_t num;
+
+		rc = parse_size(tk, (uintmax_t *) &num, &pow);
+		if (!rc) {
+			if (pow) { /* specified as <size><suffix> */
+				if (!dp->cxt->sector_size) {
+					rc = -EINVAL;
+					goto done;
+				}
+				num /= dp->cxt->sector_size;
+			} else {
+				/* specified as number of sectors */
+				fdisk_partition_size_explicit(pa, 1);
+				rc = recount_script2device_sectors(dp, &num);
+				if (rc) {
+					fdisk_warnx(dp->cxt, _("Can't recalculate partition size to the device sectors"));
+					goto done;
+				}
+			}
+
+			fdisk_partition_set_size(pa, num);
+			pa->resize = sign == '-' ? FDISK_RESIZE_REDUCE :
+				     sign == '+' ? FDISK_RESIZE_ENLARGE :
+						   FDISK_RESIZE_NONE;
+		}
+		fdisk_partition_end_follow_default(pa, 0);
+	}
+
+done:
+	DBG(SCRIPT, ul_debugobj(dp, "  size parse result: rc=%d, move=%s, size=%ju, default=%s",
+				rc, pa->resize == FDISK_RESIZE_REDUCE ? "reduce" :
+				    pa->resize == FDISK_RESIZE_ENLARGE ? "enlarge" : "none",
+				    pa->size,
+				    pa->end_follow_default ? "on" : "off"));
+	return rc;
+}
+
+
+#define FDISK_SCRIPT_PARTTYPE_PARSE_FLAGS \
+	(FDISK_PARTTYPE_PARSE_DATA | FDISK_PARTTYPE_PARSE_DATALAST | \
+	 FDISK_PARTTYPE_PARSE_SHORTCUT | FDISK_PARTTYPE_PARSE_ALIAS | \
+	 FDISK_PARTTYPE_PARSE_NAME | \
+	 FDISK_PARTTYPE_PARSE_DEPRECATED)
+
 /* dump format
  * <device>: start=<num>, size=<num>, type=<string>, ...
  */
@@ -869,11 +1129,11 @@ static int parse_line_nameval(struct fdisk_script *dp, char *s)
 	char *p, *x;
 	struct fdisk_partition *pa;
 	int rc = 0;
-	uint64_t num;
 	int pno;
 
 	assert(dp);
 	assert(s);
+	assert(dp->table);
 
 	DBG(SCRIPT, ul_debugobj(dp, "   parse script line: '%s'", s));
 
@@ -906,28 +1166,12 @@ static int parse_line_nameval(struct fdisk_script *dp, char *s)
 		p = (char *) skip_blank(p);
 
 		if (!strncasecmp(p, "start=", 6)) {
-			int pow = 0;
 			p += 6;
-			rc = next_number(&p, &num, &pow);
-			if (!rc) {
-				if (pow)	/* specified as <num><suffix> */
-					num /= dp->cxt->sector_size;
-				fdisk_partition_set_start(pa, num);
-				fdisk_partition_start_follow_default(pa, 0);
-			}
-		} else if (!strncasecmp(p, "size=", 5)) {
-			int pow = 0;
+			rc = parse_start_value(dp, pa, &p);
 
+		} else if (!strncasecmp(p, "size=", 5)) {
 			p += 5;
-			rc = next_number(&p, &num, &pow);
-			if (!rc) {
-				if (pow)	/* specified as <num><suffix> */
-					num /= dp->cxt->sector_size;
-				else		/* specified as number of sectors */
-					fdisk_partition_size_explicit(pa, 1);
-				fdisk_partition_set_size(pa, num);
-				fdisk_partition_end_follow_default(pa, 0);
-			}
+			rc = parse_size_value(dp, pa, &p);
 
 		} else if (!strncasecmp(p, "bootable", 8)) {
 			/* we use next_token() to skip possible extra space */
@@ -939,36 +1183,38 @@ static int parse_line_nameval(struct fdisk_script *dp, char *s)
 
 		} else if (!strncasecmp(p, "attrs=", 6)) {
 			p += 6;
+			free(pa->attrs);
 			rc = next_string(&p, &pa->attrs);
 
 		} else if (!strncasecmp(p, "uuid=", 5)) {
 			p += 5;
+			free(pa->uuid);
 			rc = next_string(&p, &pa->uuid);
 
 		} else if (!strncasecmp(p, "name=", 5)) {
 			p += 5;
+			free(pa->name);
 			rc = next_string(&p, &pa->name);
+			if (!rc)
+				unhexmangle_string(pa->name);
 
 		} else if (!strncasecmp(p, "type=", 5) ||
 			   !strncasecmp(p, "Id=", 3)) {		/* backward compatibility */
-			char *type;
+			char *type = NULL;
+
+			fdisk_unref_parttype(pa->type);
+			pa->type = NULL;
 
 			p += ((*p == 'I' || *p == 'i') ? 3 : 5); /* "Id=", "type=" */
 
 			rc = next_string(&p, &type);
-			if (rc)
-				break;
-			pa->type = fdisk_label_parse_parttype(
-					script_get_label(dp), type);
-			free(type);
-
-			if (!pa->type) {
-				rc = -EINVAL;
-				fdisk_unref_parttype(pa->type);
-				pa->type = NULL;
-				break;
+			if (rc == 0) {
+				pa->type = fdisk_label_advparse_parttype(script_get_label(dp),
+					type, FDISK_SCRIPT_PARTTYPE_PARSE_FLAGS);
+				if (!pa->type)
+					rc = -EINVAL;
 			}
-
+			free(type);
 		} else {
 			DBG(SCRIPT, ul_debugobj(dp, "script parse error: unknown field '%s'", p));
 			rc = -EINVAL;
@@ -985,76 +1231,20 @@ static int parse_line_nameval(struct fdisk_script *dp, char *s)
 	return rc;
 }
 
-/* original sfdisk supports partition types shortcuts like 'L' = Linux native
- */
-static struct fdisk_parttype *translate_type_shortcuts(struct fdisk_script *dp, char *str)
-{
-	struct fdisk_label *lb;
-	const char *type = NULL;
-
-	if (strlen(str) != 1)
-		return NULL;
-
-	lb = script_get_label(dp);
-	if (!lb)
-		return NULL;
-
-	if (lb->id == FDISK_DISKLABEL_DOS) {
-		switch (*str) {
-		case 'L':	/* Linux */
-			type = "83";
-			break;
-		case 'S':	/* Swap */
-			type = "82";
-			break;
-		case 'E':	/* Dos extended */
-			type = "05";
-			break;
-		case 'X':	/* Linux extended */
-			type = "85";
-			break;
-		case 'U':	/* UEFI system */
-			type = "EF";
-			break;
-		}
-	} else if (lb->id == FDISK_DISKLABEL_GPT) {
-		switch (*str) {
-		case 'L':	/* Linux */
-			type = "0FC63DAF-8483-4772-8E79-3D69D8477DE4";
-			break;
-		case 'S':	/* Swap */
-			type = "0657FD6D-A4AB-43C4-84E5-0933C84B4F4F";
-			break;
-		case 'H':	/* Home */
-			type = "933AC7E1-2EB4-4F13-B844-0E14E2AEF915";
-			break;
-		case 'U':	/* UEFI system */
-			type = "C12A7328-F81F-11D2-BA4B-00A0C93EC93B";
-			break;
-		}
-	}
-
-	return type ? fdisk_label_parse_parttype(lb, type) : NULL;
-}
-
-#define TK_PLUS		1
-#define TK_MINUS	-1
-
-#define alone_sign(_sign, _p)	(_sign && (*_p == '\0' || isblank(*_p)))
-
 /* simple format:
  * <start>, <size>, <type>, <bootable>, ...
  */
 static int parse_line_valcommas(struct fdisk_script *dp, char *s)
 {
 	int rc = 0;
-	char *p = s, *str;
+	char *p = s;
 	struct fdisk_partition *pa;
 	enum { ITEM_START, ITEM_SIZE, ITEM_TYPE, ITEM_BOOTABLE };
 	int item = -1;
 
 	assert(dp);
 	assert(s);
+	assert(dp->table);
 
 	pa = fdisk_new_partition();
 	if (!pa)
@@ -1065,84 +1255,42 @@ static int parse_line_valcommas(struct fdisk_script *dp, char *s)
 	fdisk_partition_partno_follow_default(pa, 1);
 
 	while (rc == 0 && p && *p) {
-		uint64_t num;
 		char *begin;
-		int sign = 0;
 
 		p = (char *) skip_blank(p);
 		item++;
 
-		if (item != ITEM_BOOTABLE) {
-			sign = *p == '-' ? TK_MINUS : *p == '+' ? TK_PLUS : 0;
-			if (sign)
-				p++;
-		}
-
 		DBG(SCRIPT, ul_debugobj(dp, " parsing item %d ('%s')", item, p));
 		begin = p;
 
 		switch (item) {
 		case ITEM_START:
-			if (*p == ',' || *p == ';' || alone_sign(sign, p))
-				fdisk_partition_start_follow_default(pa, 1);
-			else {
-				int pow = 0;
-
-				rc = next_number(&p, &num, &pow);
-				if (!rc) {
-					if (pow)	/* specified as <num><suffix> */
-						num /= dp->cxt->sector_size;
-					fdisk_partition_set_start(pa, num);
-					pa->movestart = sign == TK_MINUS ? FDISK_MOVE_DOWN :
-							sign == TK_PLUS  ? FDISK_MOVE_UP :
-							FDISK_MOVE_NONE;
-				}
-				fdisk_partition_start_follow_default(pa, 0);
-			}
+			rc = parse_start_value(dp, pa, &p);
 			break;
 		case ITEM_SIZE:
-			if (*p == ',' || *p == ';' || alone_sign(sign, p)) {
-				fdisk_partition_end_follow_default(pa, 1);
-				if (sign == TK_PLUS)
-					/* '+' alone means use all possible space, '-' alone means nothing */
-					pa->resize = FDISK_RESIZE_ENLARGE;
-			} else {
-				int pow = 0;
-				rc = next_number(&p, &num, &pow);
-				if (!rc) {
-					if (pow) /* specified as <size><suffix> */
-						num /= dp->cxt->sector_size;
-					else	 /* specified as number of sectors */
-						fdisk_partition_size_explicit(pa, 1);
-					fdisk_partition_set_size(pa, num);
-					pa->resize = sign == TK_MINUS ? FDISK_RESIZE_REDUCE :
-						     sign == TK_PLUS  ? FDISK_RESIZE_ENLARGE :
-							FDISK_RESIZE_NONE;
-				}
-				fdisk_partition_end_follow_default(pa, 0);
-			}
+			rc = parse_size_value(dp, pa, &p);
 			break;
 		case ITEM_TYPE:
-			if (*p == ',' || *p == ';' || alone_sign(sign, p))
+		{
+			char *str = NULL;
+
+			fdisk_unref_parttype(pa->type);
+			pa->type = NULL;
+
+			if (*p == ',' || *p == ';' || is_default_value(&p))
 				break;	/* use default type */
 
 			rc = next_string(&p, &str);
 			if (rc)
 				break;
 
-			pa->type = translate_type_shortcuts(dp, str);
-			if (!pa->type)
-				pa->type = fdisk_label_parse_parttype(
-						script_get_label(dp), str);
+			pa->type = fdisk_label_advparse_parttype(script_get_label(dp),
+						str, FDISK_SCRIPT_PARTTYPE_PARSE_FLAGS);
 			free(str);
-
-			if (!pa->type) {
+			if (!pa->type)
 				rc = -EINVAL;
-				fdisk_unref_parttype(pa->type);
-				pa->type = NULL;
-				break;
-			}
 			break;
+		}
 		case ITEM_BOOTABLE:
 			if (*p == ',' || *p == ';')
 				break;
@@ -1189,11 +1337,8 @@ static int fdisk_script_read_buffer(struct fdisk_script *dp, char *s)
 	if (!s || !*s)
 		return 0;	/* nothing baby, ignore */
 
-	if (!dp->table) {
-		dp->table = fdisk_new_table();
-		if (!dp->table)
-			return -ENOMEM;
-	}
+	if (!dp->table && fdisk_script_get_table(dp) == NULL)
+		return -ENOMEM;
 
 	/* parse header lines only if no partition specified yet */
 	if (fdisk_table_is_empty(dp->table) && is_header_line(s))
@@ -1242,7 +1387,8 @@ int fdisk_script_set_fgets(struct fdisk_script *dp,
  *
  * Reads next line into dump.
  *
- * Returns: 0 on success, <0 on error, 1 when nothing to read.
+ * Returns: 0 on success, <0 on error, 1 when nothing to read. For unknown headers
+ *          returns -ENOTSUP, it's usually safe to ignore this error.
  */
 int fdisk_script_read_line(struct fdisk_script *dp, FILE *f, char *buf, size_t bufsz)
 {
@@ -1250,11 +1396,13 @@ int fdisk_script_read_line(struct fdisk_script *dp, FILE *f, char *buf, size_t b
 
 	assert(dp);
 	assert(f);
+	assert(bufsz);
 
 	DBG(SCRIPT, ul_debugobj(dp, " parsing line %zu", dp->nlines));
 
 	/* read the next non-blank non-comment line */
 	do {
+		buf[0] = '\0';
 		if (dp->fn_fgets) {
 			if (dp->fn_fgets(dp, buf, bufsz, f) == NULL)
 				return 1;
@@ -1297,7 +1445,7 @@ int fdisk_script_read_line(struct fdisk_script *dp, FILE *f, char *buf, size_t b
  */
 int fdisk_script_read_file(struct fdisk_script *dp, FILE *f)
 {
-	char buf[BUFSIZ];
+	char buf[BUFSIZ] = { '\0' };
 	int rc = 1;
 
 	assert(dp);
@@ -1307,7 +1455,7 @@ int fdisk_script_read_file(struct fdisk_script *dp, FILE *f)
 
 	while (!feof(f)) {
 		rc = fdisk_script_read_line(dp, f, buf, sizeof(buf));
-		if (rc)
+		if (rc && rc != -ENOTSUP)
 			break;
 	}
 
@@ -1335,6 +1483,9 @@ int fdisk_script_read_file(struct fdisk_script *dp, FILE *f)
  * fdisk_new_script()). This context may be completely independent on
  * context used for fdisk_set_script().
  *
+ * Don't forget to call fdisk_set_script(cxt, NULL); to remove this reference
+ * if no more necessary!
+ *
  * Returns: <0 on error, 0 on success.
  */
 int fdisk_set_script(struct fdisk_context *cxt, struct fdisk_script *dp)
@@ -1389,6 +1540,39 @@ int fdisk_apply_script_headers(struct fdisk_context *cxt, struct fdisk_script *d
 	DBG(SCRIPT, ul_debugobj(dp, "applying script headers"));
 	fdisk_set_script(cxt, dp);
 
+	if (dp->sector_size && dp->cxt->sector_size != dp->sector_size) {
+		/*
+		 * Ignore last and first LBA if device sector size mismatch
+		 * with sector size in script.  It would be possible to
+		 * recalculate it, but for GPT it will not work in some cases
+		 * as these offsets are calculated by relative number of
+		 * sectors. It's better to use library defaults than try
+		 * to be smart ...
+		 */
+		if (fdisk_script_get_header(dp, "first-lba")) {
+			fdisk_script_set_header(dp, "first-lba", NULL);
+			fdisk_info(dp->cxt, _("Ignore \"first-lba\" header due to sector size mismatch."));
+		}
+		if (fdisk_script_get_header(dp, "last-lba")) {
+			fdisk_script_set_header(dp, "last-lba", NULL);
+			fdisk_info(dp->cxt, _("Ignore \"last-lba\" header due to sector size mismatch."));
+		}
+	}
+
+	str = fdisk_script_get_header(dp, "grain");
+	if (str) {
+		uintmax_t sz;
+
+		rc = parse_size(str, &sz, NULL);
+		if (rc == 0)
+			rc = fdisk_save_user_grain(cxt, sz);
+		if (rc)
+			return rc;
+	}
+
+	if (fdisk_has_user_device_properties(cxt))
+		fdisk_apply_user_device_properties(cxt);
+
 	/* create empty label */
 	name = fdisk_script_get_header(dp, "label");
 	if (!name)
@@ -1447,9 +1631,50 @@ int fdisk_apply_script(struct fdisk_context *cxt, struct fdisk_script *dp)
 	return rc;
 }
 
-#ifdef TEST_PROGRAM
-static int test_dump(struct fdisk_test *ts, int argc, char *argv[])
+#ifdef FUZZ_TARGET
+# include "all-io.h"
+
+int LLVMFuzzerTestOneInput(const uint8_t *data, size_t size)
 {
+	char name[] = "/tmp/test-script-fuzz.XXXXXX";
+	int fd;
+	struct fdisk_script *dp;
+	struct fdisk_context *cxt;
+	FILE *f;
+
+	fd = mkstemp_cloexec(name);
+	if (fd < 0)
+		err(EXIT_FAILURE, "mkstemp() failed");
+	if (write_all(fd, data, size) != 0)
+		err(EXIT_FAILURE, "write() failed");
+	f = fopen(name, "r");
+	if (!f)
+		err(EXIT_FAILURE, "cannot open %s", name);
+
+	cxt = fdisk_new_context();
+	dp = fdisk_new_script(cxt);
+
+	fdisk_script_read_file(dp, f);
+	fclose(f);
+
+	fdisk_script_write_file(dp, stdout);
+	fdisk_unref_script(dp);
+	fdisk_unref_context(cxt);
+
+	close(fd);
+	unlink(name);
+
+	return 0;
+}
+#endif
+
+#ifdef TEST_PROGRAM
+static int test_dump(struct fdisk_test *ts __attribute__((unused)),
+		     int argc, char *argv[])
+{
+	if (argc != 2)
+		return -1;
+
 	char *devname = argv[1];
 	struct fdisk_context *cxt;
 	struct fdisk_script *dp;
@@ -1467,8 +1692,12 @@ static int test_dump(struct fdisk_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-static int test_read(struct fdisk_test *ts, int argc, char *argv[])
+static int test_read(struct fdisk_test *ts __attribute__((unused)),
+		     int argc, char *argv[])
 {
+	if (argc != 2)
+		return -1;
+
 	char *filename = argv[1];
 	struct fdisk_script *dp;
 	struct fdisk_context *cxt;
@@ -1490,9 +1719,13 @@ static int test_read(struct fdisk_test *ts, int argc, char *argv[])
 	return 0;
 }
 
-static int test_stdin(struct fdisk_test *ts, int argc, char *argv[])
+static int test_stdin(struct fdisk_test *ts __attribute__((unused)),
+		      int argc, char *argv[] __attribute__((unused)))
 {
-	char buf[BUFSIZ];
+	if (argc != 1)
+		return -1;
+
+	char buf[BUFSIZ] = { '\0' };
 	struct fdisk_script *dp;
 	struct fdisk_context *cxt;
 	int rc = 0;
@@ -1504,7 +1737,7 @@ static int test_stdin(struct fdisk_test *ts, int argc, char *argv[])
 	printf("<start>, <size>, <type>, <bootable: *|->\n");
 	do {
 		struct fdisk_partition *pa;
-		size_t n = fdisk_table_get_nents(dp->table);
+		size_t n = dp->table ? fdisk_table_get_nents(dp->table) : 0;
 
 		printf(" #%zu :\n", n + 1);
 		rc = fdisk_script_read_line(dp, stdin, buf, sizeof(buf));
@@ -1525,8 +1758,12 @@ static int test_stdin(struct fdisk_test *ts, int argc, char *argv[])
 	return rc;
 }
 
-static int test_apply(struct fdisk_test *ts, int argc, char *argv[])
+static int test_apply(struct fdisk_test *ts __attribute__((unused)),
+		      int argc, char *argv[])
 {
+	if (argc != 3)
+		return -1;
+
 	char *devname = argv[1], *scriptname = argv[2];
 	struct fdisk_context *cxt;
 	struct fdisk_script *dp;
@@ -1567,8 +1804,12 @@ done:
 	return 0;
 }
 
-static int test_tokens(struct fdisk_test *ts, int argc, char *argv[])
+static int test_tokens(struct fdisk_test *ts __attribute__((unused)),
+		       int argc, char *argv[])
 {
+	if (argc != 2)
+		return -1;
+
 	char *p, *str = argc == 2 ? strdup(argv[1]) : NULL;
 	int i;
 
